from .client.collection_management import *
from .client.admin_management import *
from .client_models.collection import *
from .client_models.collection_builder import *
from .client_models.vector import *
from .client_models.search import *
from .client_models.column_type import *

from .utils.exceptions import (
    CollectionNotFound,
    CollectionAlreadyExists,
    UnAuthorized,
    LimitExceeded,
    ServerError,
    APIError,
    Unimplemented,
)



import os
from typing import List, Dict
from urllib import request

from pydantic import BaseModel

from nautilusdb.client_models.column_type import ColumnType
from nautilusdb.client_models.app import AnswerReference
from nautilusdb.client_models.search import SearchRequest, SearchResponse
from nautilusdb.server_models.app_api import AddDocRequest, AskRequest, AskResponse
from nautilusdb.server_models.search_api import (
    SearchWithEmbedding,
    SearchRequest as ServerQueryRequest,
    SearchResponse as ServerQueryResponse,
)
from nautilusdb.utils.config import Config
from nautilusdb.utils.exceptions import Unimplemented
from nautilusdb.client_models.vector import Vector
from nautilusdb.server_models.vector_api import (
    Vector as ServerVector,
    UpsertRequest,
    UpsertResponse,
)
from nautilusdb.utils.file_upload import (
    is_url,
    validate_file,
    validate_url_file,
)

__all__ = ["Collection"]


class Collection(BaseModel):
    """
    A Collection is a named vector search index with a fixed embedding dimension.

    Usage:
        import nautilusdb as ndb
        collection =  ndb.create_collection(ndb.
    """
    #  Name of the collection
    name: str

    # Dimension of the vectors in this collection
    dimension: int = 0

    # Description of the collection
    description: str = ""

    # A set of metadata columns associated with vectors in this collection.
    #       Key: Column name
    #       Value: Column type(avro primitive type)
    metadata_columns: Dict[str, ColumnType] = {}

    def upsert_vector(self, vectors: List[Vector]) -> int:
        """
        Upserts a list of vectors.

        For every vector in the list:
        - If a vector with the same ID already exists in the collection,
          it will be replaced with the provided vector
        - If a vector with the same ID does not exist in the collection, the
          provided vector will be created.

        :param vectors: A list of vectors to create or update
        :return: The number of vectors created or updated
        """
        url = f'{Config.get_base_url()}/vectors/upsert'
        req = UpsertRequest(
            collection_name=self.name,
            vectors=[ServerVector.from_client_vector(v) for v in vectors])
        resp = Config.post(url=url, data=req.model_dump_json())
        resp = UpsertResponse.model_validate(resp.json())
        return resp.upsert_count

    def delete_vector(self, vector_ids=List[str]):
        """
        Deletes vectors specified by the given list of vector IDs from the
        collection.

        :param vector_ids: IDs of vectors to delete
        :return:
        """
        raise Unimplemented("delete_vector is not yet implemented")

    def upload_document(self, path_or_url: str):
        """
        Uploads a local or web document into this collection.

        :param path_or_url: Local file path or the URL of a document. URL must
                            contain scheme (http or https) prefix.
        :return:
        """
        url = f'{Config.get_base_url()}/qadocs/add'
        req = AddDocRequest(collection_name=self.name)
        if is_url(path_or_url):
            validate_url_file(path_or_url)
            with request.urlopen(path_or_url) as f:
                fname = os.path.basename(path_or_url)
                Config.post(
                    url=url, files={'file': (fname, f)},
                    data={'request': req.model_dump_json()})
        else:
            # Handle tilde "~/blah"
            path_or_url = os.path.expanduser(path_or_url)
            validate_file(path_or_url)
            with open(path_or_url, 'rb') as f:
                fname = os.path.basename(path_or_url)
                Config.post(
                    url=url, files={'file': (fname, f)},
                    data={'request': req.model_dump_json()})

    def ask(
        self,
        question: str,
    ) -> (str, List[AnswerReference]):
        """
        Ask a question within the context of this collection.

        :param question: Question in plain text.
        :return: A tuple
            - answer: A plain-text answer to the given question
            - references: A list of most relevant document references in the
                          collection
        """
        url = f'{Config.get_base_url()}/qadocs/ask'
        req = AskRequest(collection_name=self.name, question=question)

        resp = Config.post(url=url, data=req.model_dump_json())
        resp = AskResponse.model_validate(resp.json())

        unique_doc_names = {ref.doc_name for ref in resp.refs}

        return (
            resp.answer,
            [AnswerReference(doc_name=name) for name in unique_doc_names]
        )

    def search(self, queries: List[SearchRequest]) -> List[SearchResponse]:
        """
        Searches the collection
        """

        req = ServerQueryRequest(
            collection_name=self.name, queries=[
                SearchWithEmbedding.from_client_request(q) for q in queries])

        url = f'{Config.get_base_url()}/vectors/search'

        resp = Config.post(url=url, data=req.model_dump_json())
        resp = ServerQueryResponse.model_validate(resp.json())
        return [r.to_client_response() for r in resp.results]



import unittest
import uuid

import nautilusdb as ndb


class SmokeTest(unittest.TestCase):
    """
    Verifies basic functionality of client. Not a hermetic test, it invokes real
    nautilusdb APIs
    """
    apikey1: str
    apikey2: str

    def setUp(self):
        # TODO: Replace hard-coded API key with dynamically created keys when
        # TODO: we have the ability to delete API keys
        #self.apikey1 = ndb.create_api_key()
        #self.apikey2 = ndb.create_api_key()
        self.apikey1 = 'MS1hOTJmZDE1Ni1lNTg1LTcyM2ItMzZiNy0yYjEyYzdjZDQ3ZWE='
        self.apikey2 = 'MS1iMjA0ZDc1Yi03MTc5LTZlMTgtMjBmMC02OWQzODZiOTExZDM='

    def test_invalid_api_key(self):
        ndb.init(api_key='invalid')

        # Invalid api_key triggers exception
        self.assertRaises(ndb.UnAuthorized, ndb.list_collections)

    def test_collection_apis(self):
        # Clear API key from config
        ndb.init(api_key=None)

        apikey1 = self.apikey1
        apikey2 = self.apikey2

        # Create a public collection
        public_col = self.create_collection()

        # Create private collections, one for each created key
        ndb.init(api_key=apikey1)
        private_col_key1 = self.create_collection()

        ndb.init(api_key=apikey2)
        private_col_key2 = self.create_collection()

        try:
            # List collections without API key
            ndb.init(api_key=None)
            collections_visible_to_public = set(ndb.list_collections())
            assert private_col_key1.name not in collections_visible_to_public
            assert private_col_key2.name not in collections_visible_to_public
            assert public_col.name in collections_visible_to_public

            # List collections using API key1
            ndb.init(api_key=apikey1)
            collections_visible_to_key1 = set(ndb.list_collections())
            assert private_col_key2.name not in collections_visible_to_key1
            assert private_col_key1.name in collections_visible_to_key1
            assert public_col.name in collections_visible_to_key1
            # Key1 is not authorized to delete a collection created by key2
            self.assertRaises(ndb.UnAuthorized, ndb.delete_collection, private_col_key2.name)

            # List collections using API key2
            ndb.init(api_key=apikey2)
            collections_visible_to_key2 = set(ndb.list_collections())
            assert private_col_key1.name not in collections_visible_to_key2
            assert private_col_key2.name in collections_visible_to_key2
            assert public_col.name in collections_visible_to_key2
            # Key2 is not authorized to delete a collection created by key1
            self.assertRaises(ndb.UnAuthorized, ndb.delete_collection, private_col_key1.name)

        finally:
            # key1 is authorized to delete its own collections as well as public
            # collections
            ndb.init(api_key=apikey1)
            ndb.delete_collection(public_col.name)
            ndb.delete_collection(private_col_key1.name)

            ndb.init(api_key=apikey2)
            ndb.delete_collection(private_col_key2.name)

    def create_collection(self) -> ndb.Collection:
        unique_col = uuid.uuid4().hex
        col = (
            ndb.CollectionBuilder.question_answer(unique_col).build())
        return ndb.create_collection(col)

    def test_vector_apis(self):
        owner = self.apikey1
        ndb.init(owner)
        unique_col = uuid.uuid4().hex
        col = ndb.CollectionBuilder().set_name(unique_col).set_dimension(2).build()
        ndb.create_collection(col)

        try:
            col.upsert_vector([ndb.Vector(vid="abc123", embedding=[1.1, 2.2])])
            col.upsert_vector([ndb.Vector(vid="abc", embedding=[1.1, 2.2])])
            col.upsert_vector(
                [
                    ndb.Vector(vid="123", embedding=[1.1, 2.2]),
                    ndb.Vector(vid="124", embedding=[1.1, 2.2]),
                    ndb.Vector(vid="125", embedding=[1.1, 2.2]),
                ])

            # Upsert is idempotent
            col.upsert_vector(
                [
                    ndb.Vector(vid="123", embedding=[1.1, 2.2]),
                    ndb.Vector(vid="124", embedding=[1.1, 2.2]),
                    ndb.Vector(vid="125", embedding=[1.1, 2.2]),
                ])
        finally:
            # Delete is idempotent
            ndb.delete_collection(col.name)

    def test_search_metadata_filter(self):
        owner = self.apikey1
        ndb.init(owner)
        unique_col = uuid.uuid4().hex

        # Create a collection with a few different metadata columns
        col = (ndb.CollectionBuilder()
               .set_name(unique_col)
               .set_dimension(2)
               .add_metadata_column('int_column', ndb.ColumnType.Int)
               .add_metadata_column('float_column', ndb.ColumnType.Float)
               .add_metadata_column('double_column', ndb.ColumnType.Double)
               .add_metadata_column('string_column', ndb.ColumnType.String)
               .add_metadata_column('bytes_column', ndb.ColumnType.Bytes)
               .add_metadata_column('boolean_column', ndb.ColumnType.Boolean)
               .build())
        ndb.create_collection(col)

        try:
            # Vector 1-3 are at (1, 1) - (3, 3) with all metadata columns populated
            # Vector 10-30 are at (10, 10) - (30, 30) with some metadata columns populated
            # Vector 100-300 are at (100, 100) - (300, 300) with no metadata columns populated
            col.upsert_vector(
                [
                    ndb.Vector(
                        vid="1",
                        embedding=[1.0, 1.0],
                        metadata={
                            'int_column': 1,
                            'float_column': 1.0,
                            'double_column': 1.0,
                            'string_column': "vector at 1.0, 1.0",
                            'boolean_column': True,
                        }),
                    ndb.Vector(
                        vid="2",
                        embedding=[2.0, 2.0],
                        metadata={
                            'int_column': 2,
                            'float_column': 2.0,
                            'double_column': 2.0,
                            'string_column': "vector at 2.0, 2.0",
                            'boolean_column': True,
                        }),
                    ndb.Vector(
                        vid="3",
                        embedding=[3.0, 3.0],
                        metadata={
                            'int_column': 3,
                            'float_column': 3.0,
                            'double_column': 3.0,
                            'string_column': "vector at 3.0, 3.0",
                            'boolean_column': True,
                        }),
                    ndb.Vector(
                        vid="10",
                        embedding=[10.0, 10.0],
                        metadata={'int_column': 10, 'float_column': 10.0, 'double_column': 10.0}),
                    ndb.Vector(
                        vid="20",
                        embedding=[20.0, 20.0],
                        metadata={'int_column': 20, 'float_column': 20.0, 'double_column': 20.0}),
                    ndb.Vector(
                        vid="30",
                        embedding=[30.0, 30.0],
                        metadata={'int_column': 30, 'float_column': 30.0, 'double_column': 30.0}),
                    ndb.Vector(vid="100", embedding=[100.0, 100.0]),
                    ndb.Vector(vid="200", embedding=[200.0, 200.0]),
                    ndb.Vector(vid="300", embedding=[300.0, 300.0]),
                ])

            # all search queries are constructed with a vector at [0.1, 0.1]. The closest vectors
            # are 1, 2, 3, 10, 20, 30, 100, 200, 300 in that order

            filter_and_expectations = {
                # Test No filter
                None: {'1', '2', '3'},

                # Test simple filter
                "int_column != 1": {'2', '3', '10'},
                "int_column = 10": {'10'},

                # Test AND
                "int_column > 1 and float_column <= 3.1": {'2', '3'},

                # Test OR
                "int_column = 1 or int_column = 2 or int_column = 3": {'1', '2', '3'},

                # Test NOT
                "int_column != 1 and not int_column = 2": {'3', '10', '20'},

                # Test is/is not null
                "int_column is null": {'100', '200', '300'},
                "int_column is not null and float_column > 3.1": {'10', '20', '30'},

                # Test arithmetics
                "int_column + 1 = 2": {'1'},
                "int_column % 2 = 0": {'2', '10', '20'},

                # Test parenthesis
                "(int_column < 100 and int_column > 10) or int_column < 2": {'1', '20', '30'},

                # Test other columns
                "string_column = 'vector at 3.0, 3.0' or boolean_column = false": {'3'},
                "boolean_column = true": {'1', '2', '3'},
            }

            requests = [ndb.SearchRequest(embedding=[0.1, 0.1], metadata_filter=ftr) for ftr
                        in filter_and_expectations.keys()]
            results = col.search(requests)
            actual_ids = [{vector.vid for vector in result.vectors} for result in results]
            assert len(actual_ids) == len(filter_and_expectations)

            for i in range(len(actual_ids)):
                ftr = requests[i].metadata_filter
                expectation = filter_and_expectations[ftr]
                assert actual_ids[i] == expectation, (f"Expected {expectation}, got "
                                                      f"{actual_ids[i]} for filter {ftr}")

        finally:
            # Delete is idempotent
            ndb.delete_collection(col.name)



from typing import List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.vector import VectorResult


__all__ = ["SearchRequest", "SearchResponse"]


class SearchRequest(BaseModel):
    """
    A query against a collection.
    """

    # Embedding of the query
    embedding: List[float]

    """
    SQL-compatible Metadata filter. Filter operations can be performed on any
    metadata column declared in the collection of supported data type.
    
    Supported data types: Boolean, Int, Long , Float, Double, String
    
    Supported filter operators (a subset of SQL:1999 standard)
    - Arithmetic Operators: + , - , * , / , %
    - Comparison Operators: =, <, >, <=, >=, !=
    - Boolean Operators: and, or, not
    - Grouping Operators: ()
    - Null Check: is null, is not null
    
    Note about filter syntax:
    - Do not enclose metadata column names
    - Do not enclose Int, Long, Float or Double values.
    - Enclose String values with single quote
   
    Examples:
    1. A Collection contains 'published_on (int)' metadata column that contains
       the epoch timestamp of when a vector is published. 'tag (string)' column 
       that contains a tag for the vector. The following filter can be used to
       restricts query results to only important vectors published after 
       1/1/2022 GMT.
    
       published_on > 1641024000 and tag = 'important'
       
    2. A Collection contains reddit comments with metadata columns 'karma (int)'
       . 'published_on (int)', 'char_count (int)'. The following could be a 
       reasonable filter to restrict results to only recent, popular and 
       informative comments.
       
       karma >= 50 and published_on > 1641024000  and char_count > 100
       
    3. Other examples of valid filters
       (foo_int > 10 or bar_string = 'front_page') and baz_string != 'draft'
       (a = 1 or (b = 2 and c = 3 and (d = 4 or e = 5)))
    """
    metadata_filter: Optional[str] = None

    # A list of metadata columns to return with the vector embedding
    include_metadata: Optional[List[str]] = None

    # Whether embeddings are returned
    include_values: Optional[bool] = False

    # Number of results to return
    top_k: Optional[int] = 3


class SearchResponse(BaseModel):
    """
    Result of a query
    """
    vectors: List[VectorResult]



from typing import Dict, Any, List, Optional

from pydantic import BaseModel

__all__ = ["Vector", "VectorResult"]


class Vector(BaseModel):
    """
    A Vector is a document with embeddings and an optional set of key-value
    properties (metadata).
    """

    # ID of the vector
    vid: str

    # Embedding of the vector.
    #
    # Embedding dimension must be identical to the embedding dimension
    # configured for its collection.
    embedding: Optional[List[float]] = None

    # Metadata associated with this vector.
    metadata: Optional[Dict[str, Any]] = None


class VectorResult(Vector):
    """
    Vector with a numeric relevance score.
    """

    # A numeric score to denote the relative relevance of this vector to the
    # given query. Higher score indicates more relevant vector.
    # Score is currently set to 1/(1 + l2_dist) between the vector and query
    # vector.
    score: float




import math
import unittest
from typing import List

import requests_mock

import nautilusdb as ndb
from nautilusdb import CollectionBuilder, Vector
from nautilusdb.client_models.search import SearchRequest, SearchResponse
from nautilusdb.server_models.collection_api import (
    ListCollectionsResponse,
)
from nautilusdb.server_models.vector_api import (
    UpsertResponse,
    VectorWithScore as ServerVectorWithScore,
)
from nautilusdb.server_models.app_api import AskResponse, AnswerReference
from nautilusdb.server_models.search_api import (
    SearchRequest as ServerQueryRequest,
    SearchResponse as ServerQueryResponse,
    SearchResult as ServerQueryResult,
)


class APISurfaceTest(unittest.TestCase):
    API_KEY = 'fakeapikey'
    API_ENDPOINT = 'https://public.us-west-2.aws.nautilusdb.com'

    def setUp(self):
        ndb.init(api_key=APISurfaceTest.API_KEY)

    @requests_mock.mock()
    def test_create_api_key(self, mock):
        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/apikey/create',
            text='{"api_key": "anotherfakeapikey"}',
        )

        assert ndb.create_api_key() == 'anotherfakeapikey'

    @requests_mock.mock()
    def test_create_collection(self, mock):
        def match_request_text(request):
            return request.text == (
                '{"name":"fakename","dimension":10,'
                '"description":"descr",' '"metas":{}}')

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/collections/create',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
        )

        assert ndb.create_collection(
            CollectionBuilder()
            .set_name('fakename')
            .set_dimension(10)
            .set_description('descr').build())

    @requests_mock.mock()
    def test_create_qa_collection(self, mock):
        def match_request_text(request):
            return request.text == (
                '{"name":"fakename","dimension":1536,'
                '"description":"This is a demo collection. '
                'Embeddings are generated using OpenAI ada_002 server_models",'
                '"metas":{"text":"string","tokens":"int","filename":"string"}}')

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/collections/create',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
        )

        assert ndb.create_collection(CollectionBuilder.question_answer("fakename").build())

    @requests_mock.mock()
    def test_list_collections(self, mock):
        resp = ListCollectionsResponse(names=['a', 'b', 'c'])
        mock.register_uri(
            'GET',
            f'{APISurfaceTest.API_ENDPOINT}/collections/list',
            request_headers={'api-key': 'fakeapikey'},
            text=resp.model_dump_json(),
        )

        assert set(ndb.list_collections()) == {'a', 'b', 'c'}

    @requests_mock.mock()
    def test_delete_collection(self, mock):
        def match_request_text(request):
            return request.text == '{"name":"fakename"}'

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/collections/delete',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
        )

        ndb.delete_collection('fakename')

    @requests_mock.mock()
    def test_upsert_vector(self, mock):
        def match_request_text(request):
            return request.text == (
                '{"collection_name":"fakename",'
                '"vectors":[{"id":"foo","embedding":[1.1,2.2],"metas":null}]}')

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/vectors/upsert',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
            text=UpsertResponse(upsert_count=1).model_dump_json(),
        )

        assert ndb.collection('fakename').upsert_vector(
            [Vector(vid='foo', embedding=[1.1, 2.2])]) == 1

    @requests_mock.mock()
    def test_ask(self, mock):
        def match_request_text(request):
            return request.text == (
                '{"collection_name":"fakename",'
                '"question":"what is the meaning of life?"}')

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/qadocs/ask',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
            text=AskResponse(
                answer='42',
                refs=[AnswerReference(doc_name="doc")]
            ).model_dump_json(),
        )

        answer, _ = ndb.collection('fakename').ask(
            "what is the meaning of life?")
        assert answer == '42'

    @requests_mock.mock()
    def test_search(self, mock):
        def match_request_text(request):
            req = ServerQueryRequest.model_validate_json(request.text)
            return (req.collection_name == 'fakename'
                    and len(req.queries) == 1
                    and req.queries[ 0].where == 'where clause')

        mock.register_uri(
            'POST',
            f'{APISurfaceTest.API_ENDPOINT}/vectors/search',
            request_headers={'api-key': 'fakeapikey'},
            additional_matcher=match_request_text,
            text=ServerQueryResponse(
                results=[
                    ServerQueryResult(
                        vectors=[
                            ServerVectorWithScore(score=1.0, id='vec')
                        ]
                    )
                ]
            ).model_dump_json(),
        )

        result: List[SearchResponse] = ndb.collection('fakename').search(
            [SearchRequest(embedding=[1.1, 2.2], metadata_filter='where clause')])
        assert len(result) == 1
        assert len(result[0].vectors) == 1
        assert math.isclose(result[0].vectors[0].score, 1.0)
        assert result[0].vectors[0].vid == 'vec'



from typing import List

from nautilusdb.client_models.collection import Collection
from nautilusdb.utils.config import Config

from nautilusdb.server_models.collection_api import (
    CreateCollectionRequest,
    ListCollectionsResponse,
    DeleteCollectionRequest,
)

__all__ = [
    "create_collection",
    "collection",
    "list_collections",
    "delete_collection",
]


def create_collection(col: Collection) -> Collection:
    """
    Create a new collection in NautilusDB.

    Args:
        col (Collection): The collection to create.
    Returns:
        The created collection

    """
    url = f'{Config.get_base_url()}/collections/create'
    req = CreateCollectionRequest(
        name=col.name,
        dimension=col.dimension,
        description=col.description,
        metas=col.metadata_columns)
    Config.post(url=url, data=req.model_dump_json())
    return col


def collection(collection_name: str) -> Collection:
    """
    Initializes a collection by name.

    Args:
        collection_name (str): The name of the collection.

    Returns:
        Collection: The collection object.
    """
    return Collection(name=collection_name)


def list_collections() -> List[str]:
    """
    List all collections in NautilusDB.

    Returns:
        List[Collection]: A list of collection objects.
    """
    url = f'{Config.get_base_url()}/collections/list'
    resp = Config.get(url=url)
    return ListCollectionsResponse.model_validate(resp.json()).names


def delete_collection(collection_name: str):
    """
    Delete a collection from NautilusDB.

    Args:
        collection_name (str): The name of the collection.
    """
    req = DeleteCollectionRequest(name=collection_name)
    url = f'{Config.get_base_url()}/collections/delete'
    Config.post(url=url, data=req.model_dump_json())



import json
from http import HTTPStatus

import requests

__all__ = [
    "CollectionNotFound",
    "CollectionAlreadyExists",
    "UnAuthorized",
    "LimitExceeded",
    "ServerError",
    "APIError",
    "Unimplemented",
]


class CollectionNotFound(Exception):
    pass


class CollectionAlreadyExists(Exception):
    pass


class UnAuthorized(Exception):
    pass


class LimitExceeded(Exception):
    pass


class ServerError(Exception):
    pass


class APIError(Exception):
    pass


class Unimplemented(Exception):
    pass


def _get_reason(response: requests.Response):
    try:
        reason = json.loads(response.content)
        return reason.get('detail')
    except Exception as e:
        pass
    return response.reason


def raise_for_error(response: requests.Response):
    """
    Converts and raises http errors into appropriate client-side exception

    :param response: HTTP response
    :return
    """
    match response.status_code:
        case HTTPStatus.OK:
            return
        case HTTPStatus.CONFLICT:
            raise CollectionAlreadyExists(_get_reason(response))
        case HTTPStatus.NOT_FOUND:
            raise CollectionNotFound(_get_reason(response))
        case HTTPStatus.PAYMENT_REQUIRED:
            raise LimitExceeded(_get_reason(response))
        case HTTPStatus.INTERNAL_SERVER_ERROR:
            raise ServerError(_get_reason(response))
        case HTTPStatus.UNAUTHORIZED:
            raise UnAuthorized(_get_reason(response))
        case _:
            if response.status_code >= 400:
                raise APIError(
                    f"{response.status_code}: {_get_reason(response)} "
                    f"from {response.request.url}")



import json
from typing import List

from pydantic import BaseModel


class AddDocRequest(BaseModel):
    collection_name: str

    @classmethod
    def validate_to_json(cls, value):
        return cls(**json.loads(value)) if isinstance(value, str) else value

    def to_json_string(self) -> str:
        # fastapi server expects "property name enclosed in double quotes" when
        # using with UploadFile. pydantic.model_dump_json() uses single quote.
        # explicitly uses json.dumps for AddDocRequest.
        return json.dumps(self.__dict__)


class AskRequest(BaseModel):
    collection_name: str
    question: str


class AnswerReference(BaseModel):
    doc_name: str


class AskResponse(BaseModel):
    answer: str
    refs: List[AnswerReference]


from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel

from nautilusdb.client_models.vector import (
    Vector as ClientVector,
    VectorResult as ClientVectorResult,
)


class Vector(BaseModel):
    id: str
    embedding: Optional[List[float]] = None
    metas: Optional[Dict[str, Any]] = None

    @classmethod
    def from_client_vector(cls, vector: ClientVector) -> "Vector":
        if vector.vid is None or vector.vid == "":
            raise ValueError("Cannot create vector without a name")
        if vector.embedding is None:
            raise ValueError("Cannot create vector without embeddings")
        return Vector(
            id=vector.vid,
            embedding=vector.embedding,
            metas=vector.metadata)


class VectorWithScore(Vector):
    score: float

    def to_client_vector_resut(self):
        return ClientVectorResult(
            score=self.score,
            vid=self.id,
            embedding=self.embedding,
            metadata=self.metas
        )


class UpsertRequest(BaseModel):
    collection_name: str
    vectors: List[Vector]


class UpsertResponse(BaseModel):
    upsert_count: int



from typing import Dict

from nautilusdb.client_models.collection import Collection
from nautilusdb.client_models.column_type import ColumnType

__all__ = ["CollectionBuilder"]


class CollectionBuilder:
    name: str
    dimension: int
    description: str
    metadata_columns: Dict[str, ColumnType]

    def __init__(self):
        self.name = ""
        self.dimension = 0
        self.description = ""
        self.metadata_columns = {}

    @classmethod
    def openai_ada_002(cls, name: str = "") -> "CollectionBuilder":
        """
        Canned collection config for embeddings generated with OpenAI's
        text-embedding-ada-002 model.

        :return:
        """
        return CollectionBuilder().set_dimension(1536).set_name(name)

    @classmethod
    def question_answer(cls, name: str = "") -> "CollectionBuilder":
        """
        Canned collection config that is compatible with Q/A API.

        :return:
        """
        return (CollectionBuilder()
                .set_name(name)
                .set_dimension(1536)
                .set_description('This is a demo collection. Embeddings are '
                                 'generated using OpenAI ada_002 server_models')
                .add_metadata_column('text', ColumnType.String)
                .add_metadata_column('tokens', ColumnType.Int)
                .add_metadata_column('filename', ColumnType.String))

    def set_dimension(self, dimension: int) -> "CollectionBuilder":
        self.dimension = dimension
        return self

    def set_name(self, name: str) -> "CollectionBuilder":
        self.name = name
        return self

    def set_description(self, description: str) -> "CollectionBuilder":
        self.description = description
        return self

    def add_metadata_column(
        self,
        column_name: str,
        column_type: ColumnType
        ) -> "CollectionBuilder":
        if self.metadata_columns is None:
            self.metadata_columns = {}
        self.metadata_columns[column_name] = column_type
        return self

    def build(self) -> Collection:
        if self.name is None or self.name == "":
            raise ValueError("Collection must have a name")
        if self.dimension is None or self.dimension <= 0:
            raise ValueError("Collection must have dimension > 0")
        return Collection(
            name=self.name,
            dimension=self.dimension,
            description=self.description,
            metadata_columns=self.metadata_columns)



from enum import Enum

__all__ = ["ColumnType"]


class ColumnType(str, Enum):
    """
    Metadata column data types supported by NautilusDB (avro primitive types).

    You can find the definition of each data type in
    https://avro.apache.org/docs/1.11.1/specification/#primitive-types
    """
    Boolean = "boolean"
    Int = "int"
    Long = "long"
    Float = "float"
    Double = "double"
    String = "string"
    Bytes = "bytes"



from typing import List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.search import (
    SearchRequest as ClientSearchRequest,
    SearchResponse as ClientSearchResponse,
)
from nautilusdb.server_models.vector_api import VectorWithScore


class SearchResult(BaseModel):
    vectors: List[VectorWithScore]

    def to_client_response(self) -> ClientSearchResponse:
        return ClientSearchResponse(
            vectors=[v.to_client_vector_resut() for v in self.vectors])


class Search(BaseModel):
    # where supports SQL =, <, >, <=, >=, !=, and, or, etc.
    where: Optional[str] = None
    return_metas: Optional[List[str]] = None
    return_embedding: Optional[bool] = False
    top_k: Optional[int] = 3


class SearchWithEmbedding(Search):
    embedding: List[float]

    @classmethod
    def from_client_request(
        cls,
        req: ClientSearchRequest
    ) -> "SearchWithEmbedding":
        return SearchWithEmbedding(
            embedding=req.embedding,
            where=req.metadata_filter,
            return_metas=req.include_metadata,
            return_embedding=req.include_values,
            top_k=req.top_k
        )


class SearchRequest(BaseModel):
    collection_name: str
    queries: List[SearchWithEmbedding]


class SearchResponse(BaseModel):
    # The results are sorted by the score in the descending order
    results: List[SearchResult]



from nautilusdb.server_models.admin_api import CreateApiKeyRequest, CreateApiKeyResponse
from nautilusdb.utils.config import Config

__all__ = ["init", "create_api_key"]


def init(
    api_key: str = None,
    account_name: str = Config.PUBLIC_ACCOUNT
):
    """
    Initialize the NautilusDB client.

    Args:
        api_key (str, optional): The API key to authenticate requests.
                                 Defaults to None. You can still access
                                 public collections without an API key.
        account_name (str, optional): The name of the account. Defaults to
                                      Config.DEMO_ACCOUNT
    Raises:
        ValueError: If the provided config is invalid
    """
    Config.api_key = api_key
    Config.account_name = account_name
    Config.validate_setup()


def create_api_key() -> str:
    """
    Create a new APIKey in the configured account.

    Returns:
        The created API key

    """
    url = f'{Config.get_base_url()}/apikey/create'
    req = CreateApiKeyRequest()
    resp = Config.post(url=url, data=req.model_dump_json())
    return CreateApiKeyResponse.model_validate(resp.json()).api_key



from typing import Dict, List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.column_type import ColumnType


class CreateCollectionRequest(BaseModel):
    name: str
    dimension: int
    description: Optional[str] = None
    metas: Optional[Dict[str, ColumnType]] = None


class DeleteCollectionRequest(BaseModel):
    name: str


class ListCollectionsResponse(BaseModel):
    names: List[str]





import requests

from .exceptions import raise_for_error


class Config:
    PUBLIC_ACCOUNT: str = "public"
    api_key: str = None
    account_name: str = PUBLIC_ACCOUNT

    @classmethod
    def validate_setup(cls):
        if cls.account_name is None or cls.account_name != cls.PUBLIC_ACCOUNT:
            raise ValueError("Only public account ('public') is supported.")

    @classmethod
    def get_base_url(cls):
        #return "http://localhost:8080"
        return f"https://{cls.account_name}.us-west-2.aws.nautilusdb.com"

    @classmethod
    def post(cls, url, data=None, files=None):
        resp = requests.post(
            url, data=data, files=files, headers=cls._construct_header())
        raise_for_error(resp)
        return resp

    @classmethod
    def get(cls, url, data=None):
        resp = requests.get(url, data=data, headers=cls._construct_header())
        raise_for_error(resp)
        return resp

    @classmethod
    def _construct_header(cls):
        return None if cls.api_key is None else {'api-key': cls.api_key}



import contextlib
import os
from urllib import parse

import requests


def is_url(path):
    parsed_value = parse.urlparse(path)
    return parsed_value.scheme in ("http", "https")


def validate_url_file(url):
    with contextlib.suppress(Exception):
        resp = requests.request('HEAD', url)
        nbytes = int(resp.headers.get("Content-Length"))
        if nbytes > 9_500_000:
            raise ValueError(f"File {url} exceeds size limit 9.5MB, "
                             f"actual size {nbytes} bytes")


def validate_file(path: str):
    nbyte = os.path.getsize(path)
    if nbyte > 9_500_000:
        raise ValueError(f"File {path} exceeds size limit 9.5MB, actual size "
                         f"{nbyte} bytes")